import os
import sys
from pathlib import Path
import importlib
from importlib import machinery
import inspect

import concurrent

import logging

from .redis import RedisClient
from .common import get_conf
from .event import HandlerManager

from ifconf import configure_module, config_callback

@config_callback
def config(loader):
    loader.add_attr_list('plugin_modules', [], help='module names for plugin base directory')
    loader.add_attr_int('max_workers', 1, help='max threadpool workers')

class ServerContext:

    def __init__(self, loop):
        self.logger = logging.getLogger(__name__).getChild('manager')
        self.conf = configure_module(config, override=get_conf())
        self.loop = loop
        self.session_store = {}
        self.redis = RedisClient(self.loop)
        self.event_handler_manager = HandlerManager(self)
        self.thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=self.conf.max_workers)

    def module_path(self, path):
        return path.resolve() if path.is_absolute() else Path(__file__).parent.joinpath(path).resolve()

    def resolve_local_path(self, path):
        return path.resolve() if path.is_absolute() else self.conf.ducts_home.joinpath(path).resolve()

    def plugin_paths(self, module_path, local_path):
        yield (Path(__file__).parent.joinpath(module_path), 0)
        for module_name in self.conf.plugin_modules:
            yield (Path(importlib.import_module(module_name).__file__).parent.joinpath(module_path), 1)
        yield (self.resolve_local_path(local_path), 2)

    async def run_in_executor(self, func):
        return await self.loop.run_in_executor(self.thread_pool, func)

    async def close(self):
        await self.redis.close()

    



from pathlib import Path
from ifconf import configure_module, config_callback

@config_callback
def config(loader):
    loader.add_attr_path('ducts_home', Path('.'), help='root directory of all local path')
    loader.add_attr('pidpath', '{ducts_home}/.pid/{module}_{service}.pid', help='filepath to store PID')

conf = None

def get_conf():
    global conf
    if conf:
        return conf
    conf = configure_module(config)
    return conf
    
def get_pidfile(module, service):
    global conf
    conf = get_conf()
    path = Path(conf.pidpath.format(ducts_home=conf.ducts_home, module=module, service=service))
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    return str(path.resolve().absolute())




from ducts.event import EventHandler

import asyncio

from ifconf import configure_module, config_callback

@config_callback
def config(loader):
    loader.add_attr_int('alive_monitoring_interval', 20, help='Ping interval from the server [sec]')

class Handler(EventHandler):

    def __init__(self):
        super().__init__()
        self.conf = configure_module(config)

    def setup(self, handler_spec, manager):
        handler_spec.set_description('死活監視用メインループを開始します。最初に一度だけ自動的に呼び出されます。')
        return handler_spec

    async def handle(self, event):
        yield True
        while not await event.session.is_closed():
            await asyncio.sleep(self.conf.alive_monitoring_interval)
            yield True
        
    





#!/usr/bin/env python3

import sys
import traceback
import logging

import asyncio

from ifconf import configure_module, config_callback, configure_main

from ducts.server import HttpdServer

def run():
    logger = logging.getLogger(__name__)
    loop = asyncio.get_event_loop()
    server = HttpdServer(loop)
    asyncio.ensure_future(server.run())
    try:
        loop.run_forever()
    except Exception as e:
        logger.exception('Error on loop: %s', e)
    except BaseException as e:
        logger.info(e)
    try:
        logger.info('Closing...')
        loop.run_until_complete(server.close())
    except Exception as e:
        logger.exception('Error on close: %s', e)
    except BaseException as e:
        logger.info(e)

    try:
        logger.info('Shutdown...')
        loop.run_until_complete(loop.shutdown_asyncgens())
    except Exception as e:
        logger.exception('Error on shutdown: %s', e)
    except BaseException as e:
        logger.info(e)
    finally:
        loop.close()
        logger.info('Completed. See you!')

def main():
    #configure_main()
    run()
    
if __name__ == '__main__':
    main()



import os
import sys
from pathlib import Path
import importlib
from importlib import machinery
import abc
import inspect

from collections import namedtuple
from collections import OrderedDict
from enum import Enum

from ifconf import configure_module, config_callback

import asyncio

@config_callback
def config(loader):
    loader.add_attr_path('module_path_handler', Path('./handler'), help='module path for event_handler files')
    loader.add_attr_path('local_path_handler', Path('./handler'), help='local path for event_handler files')
    loader.add_attr('plugin_filename_pattern', 'evt_*.py', help='event handler plugin filename pattern')
    loader.add_attr_int('min_user_handler_id', 1000, help='minimum handler ID for user plugin')

import logging

class EventHandler(metaclass=abc.ABCMeta):

    ATTR_KEY = 'KEY'

    @abc.abstractmethod
    def setup(self, handler_spec, manager):
        pass
    
    @abc.abstractmethod
    async def handle(self, event):
        pass
    
    def handle_closed(self, event_session):
        pass

    
class HandlerSpec():

    def __init__(self, id, key, callback):
        self.id = id
        self.key = key
        self.callback = callback
        self.responsive = False
        self.description = ''
        
        #self.input_sample = None
        #self.output_sample = None

    def __str__(self):
        return '{}({})-{}:{}'.format(type(self), id(self), self.id, self.key)
        
    def __repr__(self):
        return '{}({})-{}:{}'.format(type(self), id(self), self.id, self.key)
        
    def set_as_responsive(self):
        self.responsive = True
        
    def set_description(self, description):
        self.description = description

    def _asdict(self):
        return OrderedDict([(key,getattr(self, key))for key in ['id', 'key', 'callback', 'responsive', 'description']])


class Event(object):

    def __init__(self, event_id, event_session, data = None):
        assert event_id >= 0, 'event_id cannot be negative.'
        assert event_session, 'event_session cannot be null.'
        self.id = event_id
        self.session = event_session
        self.data = data

        
class EventSession(metaclass=abc.ABCMeta):

    def __init__(self, manager):
        self.manager = manager
        self.redis = manager.redis
        self.loop = manager.loop
        self.props = {}

    def get_property(self, key):
        return self.props[key]

    def set_property(self, key, value):
        self.props[key] = value

    @abc.abstractmethod
    async def is_closed(self):
        pass

    @abc.abstractmethod
    def request_id(self):
        pass

    @abc.abstractmethod
    def session_id(self):
        pass


class HandleType(Enum):
    ALIVE_MONITORING = 0
    ASYNC_FOR = 1
    FOR = 2
    ASYNC = 3
    SYNC = 4

    
class HandlerManager(object):

    def __init__(self, context):
        self.logger = logging.getLogger(__name__).getChild('manager')
        self.conf = configure_module(config)
        self.context = context
        self.loop = context.loop
        self.redis = context.redis
        self._helpers = {}
        self.specs = []
        self.key_ids = {}
        self._handlers = {}
        self._plugin_dirs = []
        for path, level in self.context.plugin_paths(self.conf.module_path_handler, self.conf.local_path_handler):
            self.load_plugins(path, level < 2)

    def load_plugins(self, plugin_dir, is_system_handler):
        if is_system_handler:
            self.logger.debug('SYSTEM_PLUGIN_DIR=%s', plugin_dir)
        else:
            self.logger.info('PLUGIN_DIR=%s', plugin_dir)
        self._plugin_dirs.insert(0, plugin_dir)
        local_modules = set()
        for p in plugin_dir.glob(self.conf.plugin_filename_pattern):
            loader = machinery.SourceFileLoader(p.stem, str(p))
            module = loader.load_module()
            local_modules.add(module.__name__)
        for cls in sorted(EventHandler.__subclasses__(), key=lambda x: (str(x), x)):
            h = cls()
            if h.__class__.__module__ not in local_modules:
                continue
            spec = self.setup(h, is_system_handler)
            if spec is None:
                continue
            self.specs.append(spec)
            self.key_ids[spec.key] = spec.id if not spec.key in self.key_ids else max(self.key_ids[spec.key], spec.id)
            if spec.id == 0:
                htype = HandleType.ALIVE_MONITORING
            elif inspect.isasyncgenfunction(h.handle):
                htype = HandleType.ASYNC_FOR
            elif inspect.isgeneratorfunction(h.handle):
                htype = HandleType.FOR 
            elif inspect.iscoroutinefunction(h.handle):
                htype = HandleType.ASYNC
            else:
                htype = HandleType.SYNC
            self._handlers[spec.id] = (htype, h)
        if self.logger.isEnabledFor(logging.DEBUG):
            for s in self.specs:
                self.logger.debug('HANDLER|SPEC=%s', s)
        

    def setup(self, handler, is_system_handler):
        index = 0
        for ids in handler.__module__.split('_'):
            try:
                index += (len(ids) + 1)
                id = int(ids)
                setattr(handler, 'ID', id)
                key = getattr(handler, handler.ATTR_KEY, handler.__module__[index:])
                if not key:
                    key = 'undefined'
                key = key.upper()
                setattr(handler, handler.ATTR_KEY, key)
                callback = inspect.isasyncgenfunction(handler.handle) or inspect.isgeneratorfunction(handler.handle)
                #desc = handler.description()
                if id < self.conf.min_user_handler_id and not is_system_handler:
                    self.logger.warn('Permission denied. User handler ID must be less than [%s] but was [%s]', self.conf.min_user_handler_id, id)
                    return None
                spec = HandlerSpec(id, key, callback)
                return handler.setup(spec, self)
            except ValueError:
                continue

    def load_helper_module(self, module_name):
        return self._helpers[module_name] if module_name in self._helpers else self._new_helper_module(module_name)

    def get_handler_for(self, event_id):
        return self._handlers[event_id]

    def handlers(self):
        return self._handlers.values()

    def _new_helper_module(self, module_name):
        for plugin_dir in self._plugin_dirs:
            p = plugin_dir.joinpath("{}.py".format(module_name))
            loader = machinery.SourceFileLoader(module_name, str(p))
            module = loader.load_module()
            self._helpers[module_name] = module
            return module
        
    





import base64
import binascii

from aiohttp import hdrs, web
from datetime import datetime

import inspect
import hashlib
import base64

from ifconf import configure_module, config_callback

import logging
        
@config_callback
def config(loader):
    loader.add_attr('hash_algorithm', 'sha1', help='hash algorithm name for session')
    loader.add_attr_list('white_paths', ['/libs/', '/wsd', '/main'], help='paths without authentication')

class Auth(object):

    def __init__(self, context):
        self.conf = configure_module(config)
        self.context = context
        for k,v in inspect.getmembers(hashlib, inspect.isbuiltin):
            if k == self.conf.hash_algorithm:
                self.hash_func = v
                break
        if not hasattr(self, 'hash_func'):
            raise ValueError('hash algorithm[{}] not found.'.format(self.conf.hash_algorithm_key))

    async def init_session(self, request):
        ip, port = request.transport.get_extra_info('peername')
        ua = request.headers['User-Agent']
        query = request.rel_url.query
        key = query.get('uuid', None)
        key = key if key else query.get('ts', None)
        if not key:
            raise web.HTTPForbidden()
        ssid = self.hash_func(ip.encode()+str(port).encode()+ua.encode()+key.encode()).hexdigest()
        tsid = await self.context.redis.xadd('SESSION/2020', ip=ip, port=port, ua=ua, ssid=ssid, **query)
        return "{}.{}".format(ssid, base64.b64encode(tsid).decode())

    async def check_session(self, request):
        ssid, tsids = request.match_info['session_id'].split('.')
        tsid = base64.b64decode(tsids).decode()
        ts = datetime.fromtimestamp(int(tsid.split('-')[0])/1000)
        ip, port = request.transport.get_extra_info('peername')
        ua = request.headers['User-Agent']
        param = await self.context.redis.xget_str('SESSION/2020', tsid)
        if param['ssid'] != ssid or param['ua'] != ua:
            raise web.HTTPForbidden()
        return 'SS.'+tsids
    
        #op = int(param['port'])
        #if param['ip'] != ip or port > op+128 or (op > 65535 - 128 and port > 49152 + 128):
        #    raise web.HTTPForbidden()

    async def setup(self, app):
        app.middlewares.append(self.middleware)

    async def raise_error(self, request):
        raise web.HTTPUnauthorized(
            headers={
                hdrs.WWW_AUTHENTICATE: 'Basic realm={}'.format(self._realm)
            },
        )

    @web.middleware
    async def middleware(self, request, handler):
        if request.path not in self._white_paths:
            auth_header = request.headers.get(hdrs.AUTHORIZATION)

            if auth_header is None or not auth_header.startswith('Basic '):
                return await self.raise_error(request)

            try:
                secret = auth_header[6:].encode('utf-8')

                auth_decoded = base64.decodebytes(secret).decode('utf-8')
            except (UnicodeDecodeError, UnicodeEncodeError,
                    binascii.Error):
                await self.raise_error(request)

            credentials = auth_decoded.split(':')

            if len(credentials) != 2:
                await self.raise_error(request)

            username, password = credentials

            if username != self._username or password != self._password:
                await self.raise_error(request)

        return await handler(request)



from ducts.event import EventHandler

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        handler_spec.set_description('サーバに文字列を送信し、ログとして記録します。')
        return handler_spec

    def handle(self, event):
        pass
    





from ducts.event import EventHandler

from datetime import datetime

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        handler_spec.set_description('送受信テスト用イベントです。サーバのUNIX時間を返します。')
        handler_spec.set_as_responsive()
        return handler_spec

    def handle(self, event):
        return datetime.now().strftime('%s')





#!/usr/bin/env python3
# coding: utf-8

from .app import main
    
if __name__ == '__main__':
    import ducts.server
    import ducts.redis
    import ducts.backend.asr
    
    main(config_arg = 'config_{module}.ini'
         , callback_methods = [ducts.app.config, ducts.server.config, ducts.redis.config, ducts.backend.asr.config])
    
